"""Commonly needed data & code"""

import argparse
from collections import namedtuple
import enum
import os
import re
import rich.console
import rich.progress
import textwrap


class YacceException(RuntimeError):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


# adapted from https://github.com/Arech/benchstats/blob/be9e925ae85b7dc1c19044ad5f6eddea681f9f77/src/benchstats/common.py#L56
class LoggingConsole(rich.console.Console):
    # @enum.verify(enum.CONTINUOUS)  # not supported by Py 3.10
    class LogLevel(enum.IntEnum):
        Trace = 0
        Debug = 1
        Info = 2
        Warning = 3
        Error = 4  # recoverable
        Failure = 5  # non-recoverable, can continue working
        Critical = 6  # non-recoverable, must abort

    def __init__(self, log_level: LogLevel = LogLevel.Trace, **kwargs):
        assert isinstance(log_level, LoggingConsole.LogLevel)
        self.log_level = log_level
        self._n_errors: int = 0
        if "emoji" not in kwargs:
            kwargs["emoji"] = False
        if "highlight" not in kwargs:
            kwargs["highlight"] = False
        super().__init__(**kwargs)

    def cleanNumErrors(self) -> int:
        r = self._n_errors
        self._n_errors = 0
        return r

    def getNumErrors(self) -> int:
        return self._n_errors

    def _do_log(self, color: str, lvl: str, *args, **kwargs):
        if "sep" in kwargs:
            sep = kwargs["sep"] if len(kwargs["sep"]) > 0 else " "
        else:
            sep = " "
            kwargs["sep"] = sep
        return super().print(f"[[{color}]{lvl:4s}[/{color}]]{sep}", *args, **kwargs)

    def will_log(self, level) -> bool:
        return self.log_level <= level

    def trace(self, *args, **kwargs):
        if self.log_level > LoggingConsole.LogLevel.Trace:
            return None
        return self._do_log("blue", "trce", *args, **kwargs)

    def debug(self, *args, **kwargs):
        if self.log_level > LoggingConsole.LogLevel.Debug:
            return None
        return self._do_log("bright_black", "dbg", *args, **kwargs)

    def info(self, *args, **kwargs):
        if self.log_level > LoggingConsole.LogLevel.Info:
            return None
        return self._do_log("bright_white", "info", *args, **kwargs)

    def warning(self, *args, **kwargs):
        if self.log_level > LoggingConsole.LogLevel.Warning:
            return None
        return self._do_log("yellow", "warn", *args, **kwargs)

    def error(self, *args, **kwargs):
        self._n_errors += 1
        if self.log_level > LoggingConsole.LogLevel.Error:
            return None
        return self._do_log("red", "Err", *args, **kwargs)

    def failure(self, *args, **kwargs):
        self._n_errors += 1
        if self.log_level > LoggingConsole.LogLevel.Failure:
            return None
        return self._do_log("bright_red", "FAIL", *args, **kwargs)

    def critical(self, *args, **kwargs):
        self._n_errors += 1
        if self.log_level > LoggingConsole.LogLevel.Critical:
            return None
        return self._do_log("bright_magenta", "CRIT", *args, **kwargs)

    def yacce_begin(self):
        super().print("[bold bright_blue]==== YACCE >>>>>>>>[/bold bright_blue]")

    def yacce_end(self):
        super().print("[bold bright_blue]<<<<<<<< YACCE ====[/bold bright_blue]")


kMainDescription = (
    "Yacce extracts compile_commands.json and build system insights from a build system "
    "by supervising the local compilation process with strace.\n"
    "Primarily supports Bazel (other build systems might be added later).\n"
    "--> Homepage: https://github.com/Arech/yacce"
)


# WARNING: argparse doesn't guarantee its private API stability. WTH?!
class BetterHelpFormatter(argparse.HelpFormatter):
    kMaxWidth = 100

    @staticmethod
    def _useWidth(width):
        return width if width < BetterHelpFormatter.kMaxWidth else BetterHelpFormatter.kMaxWidth

    def _fill_text(self, text, width, indent):
        width = self._useWidth(width)
        return "\n".join(
            indent + ("" if s == "%" else s)
            for line in text.splitlines()
            for s in argparse.HelpFormatter._split_lines(self, line if line else "%", width)
        )

    def _split_lines(self, text, width):
        width = self._useWidth(width)
        return [
            "" if s == "%" else s
            for line in text.splitlines()
            for s in argparse.HelpFormatter._split_lines(self, line if line else "%", width)
        ]


def addCommonCliArgs(parser: argparse.ArgumentParser, addendums: dict = {}):
    """ "Adds arguments common for multiple modes to the given parser."""
    kWarnCustomField = (
        "WARNING: current clangd gets upset when it finds a field it doesn't know, "
        "so enabling this option might prevent you from using clangd with the resulting file!\n"
    )
    kEmptyDisables = 'Pass an empty string "" to disable.'
    kAcceptSequence = "Accepts multiple values at once."

    parser.add_argument(
        "--cwd",
        help="Path to the working directory of the compilation.\n"
        "This value goes to a 'directory' field of an "
        "entry of compile_commands.json and is used to resolve relative paths found in the command. "
        "If '--ignore-not-found' argument isn't set, yacce will try to test if mentioned files exist in this "
        "directory and warn if they aren't. Note that passing the file existence test helps, but doesn't "
        "guarantee that the resulting compile_commands.json will be correct.\n"
        + addendums.get("cwd", ""),
        metavar="path/to/dir",
        # no default as it depends on the mode
    )

    parser.add_argument(
        "--ignore-not-found",
        help="If set, will not test if files to be added to .json exists.\n",
        action=argparse.BooleanOptionalAction,
        default=False,
    )

    parser.add_argument(
        "-o",
        "--other_commands",
        help="If set, yacce will also generate other_commands.json file.\n"
        "This file has a similar to compile_commands.json format, but contains all other compiler "
        "invocations found that aren't useful for gathering C++ symbol information of the project, "
        "but handy to get insights about the build in general (such as for compiling assembler "
        "sources or for linking).\n" + addendums.get("other_commands", ""),
        action=argparse.BooleanOptionalAction,
        default=False,
    )

    parser.add_argument(
        "--save_duration",
        help="If set, yacce will add a 'duration_s' field into the resulting .json that contain how "
        "long the command run in seconds with a microsecond resolution.\n"
        "This feature currently doesn't have automated use, but the file can be inspected manually, "
        "or with a custom script to obtain build system performance insights.\n" + kWarnCustomField,
        action=argparse.BooleanOptionalAction,
        default=False,
    )

    parser.add_argument(
        "--save_line_num",
        help="If set, yacce will add a 'line_num' integer field into the resulting .json that "
        "contain a line number of the compiler call in the strace log file.\n"
        "Useful for debugging, but have no automated use.\n" + kWarnCustomField,
        action=argparse.BooleanOptionalAction,
        default=False,
    )

    parser.add_argument(
        "--discard_outputs_with_pfx",
        help="A build system can compile some dummy source files only to gather information about "
        "compiler capabilities. Presence of these files in the compile_commands.json aren't usually helpful. "
        "Typically, such files are placed into /tmp or /dev/null, but other variants are possible.\n"
        "This setting allows to fully customize which prefixes of a compiler's output file should "
        "lead to ignoring the compilation call.\n" + kEmptyDisables + " " + kAcceptSequence + "\n"
        "Default: %(default)s. ",
        metavar="path/prefix",
        nargs="*",
        default=["/dev/null", "/tmp/"],
    )

    parser.add_argument(
        "--discard_sources_with_pfx",
        help="Similar to --discard_outputs_with_pfx, but controls which prefixes of source files "
        "should lead to ignoring the compiler call.\n"
        + kEmptyDisables
        + " "
        + kAcceptSequence
        + "\nDefault: %(default)s.",
        metavar="path/prefix",
        nargs="*",
        default=[""],
    )

    kDashToPlus = (
        "ATTENTION: since Python's argparse always treats a leading dash in a CLI argument as a "
        "script's argument name, but not value, use a plus sign '+' instead of a dash '-' to specify "
        "a leading dash."
    )

    parser.add_argument(
        "--discard_args_with_pfx",
        help="Certain compiler arguments, such as sanitizers, are known to choke clangd. Some others "
        "like those concerning build reproducibility might be useless for C++ symbols.\n"
        "Set a value of this parameter to a sequence of prefixes to match and remove such compiler "
        "arguments.\n"
        + kEmptyDisables
        + " "
        + kAcceptSequence
        + "\n"
        + kDashToPlus
        + " Example: instead of '-fsanitize' use '+fsanitize'.\nDefault: %(default)s.",
        metavar="+compiler_arg_prefix",
        nargs="*",
        default=[
            "+fsanitize"
        ],  # , "-frandom-seed=", "-D__DATE__=", "-D__TIMESTAMP__=", "-D__TIME__="],
    )

    parser.add_argument(
        "--discard_args",
        help="Similarly to '--discard_args_with_pfx', values for this argument define a set of "
        "compiler arguments (such as '-DMY_DEF=VALUE') or pipe-delimited argument pairs (like a "
        "single token value '-I|/certain/dir' defines a two token pair '-I /certain/dir') "
        "that will be removed from a compiler invocation.\n"
        "Note that a single token specification of a '-D' compiler argument has a special handling "
        "and also addresses its two token alternatives.\n"
        + kEmptyDisables
        + " "
        + kAcceptSequence
        + "\n"
        + kDashToPlus
        + " For the above it's '+DMY_DEF=VALUE' and '+I|/certain/dir'.\nDefault: %(default)s.",
        metavar="+compiler_arg_or_args_pair_spec",
        nargs="*",
        default=["+DADDRESS_SANITIZER"],
    )

    parser.add_argument(
        "--enable_dupes_check",
        help="If set, yacce will report if a pair <source, output> isn't unique.\n"
        "Usefulness of this flag solely depends on actual build system implementation. Some might "
        "use lots of temporary compilations just to gather compiler capabilities which could lead "
        "to an avalanche of false positives. This could be mitigated with --discard* family of flags, "
        "but this requires manual intervention, hence it's disabled by default.\n",
        action=argparse.BooleanOptionalAction,
        default=False,
    )

    parser.add_argument(
        "-c",
        "--compiler",
        help="Adds a set of absolute paths and basenames of custom compilers used by the build "
        "system to the set of compilers already detectable by yacce. " + kAcceptSequence,
        metavar="compiler_basename_or_path",
        nargs="*",
    )

    parser.add_argument(
        "-d",
        "--dest_dir",
        help="Destination directory in which yacce should create resulting .json files. Must exist.\n"
        + addendums.get("dest_dir", ""),
        metavar="dir/path",
        # no default as it depends on the mode
    )

    return parser


def warnClangdIncompatibilitiesIfAny(Con: LoggingConsole, args: argparse.Namespace) -> None:
    def _warnField(flag: str, fld_name: str):
        Con.warning(
            f"Each .json entry have a custom field '{fld_name}' that might be incompatible "
            f"with clangd. Don't use --{flag} CLI argument if you intend to feed resulting compile_commands.json to clangd."
        )

    if args.save_duration:
        _warnField("save_duration", "duration_s")

    if args.save_line_num:
        _warnField("save_line_num", "line_num")


CompilersTuple = namedtuple("CompilersTuple", ["basenames", "fullpaths"])


def makeCompilersSet(custom_compilers: list[str] | None) -> CompilersTuple:
    """Adds custom compilers to the set of known compilers to find in strace log."""

    if custom_compilers is None:
        custom_compilers = []
    assert isinstance(custom_compilers, list)
    assert all(isinstance(c, str) for c in custom_compilers)

    kGccVers = (9, 18)
    kGccPfxs = ("", "x86_64-linux-gnu-")
    kClangVers = (10, 25)

    basenames = frozenset(
        ["cc", "c++", "gcc", "g++", "clang", "clang++"]
        + [f"{pfx}gcc-{v}" for v in range(*kGccVers) for pfx in kGccPfxs]
        + [f"{pfx}g++-{v}" for v in range(*kGccVers) for pfx in kGccPfxs]
        + [f"clang-{v}" for v in range(*kClangVers)]
        + [f"clang++-{v}" for v in range(*kClangVers)]
        + [c for c in custom_compilers if c and not c.startswith("/")]
    )
    # note there's not much point to try to prune the set of basenames or full paths, as a build system
    # could reference a compiler in a custom path, so we can't detect its presence on the machine.

    paths = frozenset([c for c in custom_compilers if c and c.startswith("/")])

    return CompilersTuple(basenames=basenames, fullpaths=paths)


# line_num:int is the number (1 based index) of line in the log that spawned the process
# is_other:bool is the command is other command
# cmd_idx:int is the index of command in a corresponding list
ProcessProps = namedtuple(
    "ProcessProps", ("start_ts_us", "line_num", "is_other", "cmd_idx", "n_sources")
)
# args:list[str], output:str|None, source:str, line_num:int
CompileCommand = namedtuple("CompileCommand", ("args", "output", "source", "line_num"))
OtherCommand = namedtuple("OtherCommand", ("args", "output", "line_num"))


class BaseParser:
    kOutputArgs = ("-o", "--output")
    # List of a compiler arguments that should contain a file/dir path.
    # Note, we're interested only in args that are important for C++ symbols extraction
    # https://gcc.gnu.org/onlinedocs/gcc/Directory-Options.html
    # https://clang.llvm.org/docs/ClangCommandLineReference.html#diagnostic-options
    kArgIsPath = frozenset(
        kOutputArgs
        + (
            "-I",
            "-iquote",
            "-isystem",
            "-idirafter",
            "-isysroot",
            "--include-directory",
            "--include-directory-after--sysroot",
            "-cxx-isystem",
            "--library-directory",
            "-imacros",
            "--imacros",
            "-include",
            "--include",
        )
    )

    # List of a compiler arguments that could start with = or $SYSROOT specification.
    # https://gcc.gnu.org/onlinedocs/gcc/Directory-Options.html#index-I
    kCheckArgForSysrootSpec = frozenset((
        "-I",
        "-iquote",
        "-isystem",
        "-idirafter",
        "--include-directory",
    ))
    kSysrootSpec = ("=", "$SYSROOT")

    # list of a compiler arguments that should contain a path as a suffix.
    kPfxArgIsPath = (
        "-L",
        "--sysroot=",
        "-F",
        "--output=",
        "--embed-dir=",
        "--include-directory=",
        "-cxx-isystem",
        "-I",
        "-idirafter",
        "--include-directory-after=",
        "-imacros",
        "--imacros",
        "--imacros=",
        "-include",
        "--include",
        "--include=",
        "-iquote",
        "-isysroot",
        "-isystem",
        "-isystem-after",
        "-stdlib++-isystem",
        "--library-directory=",
    )
    r_pfx_arg_is_path = re.compile(r"^(" + "|".join(re.escape(s) for s in kPfxArgIsPath) + r")")

    # list of important for compilation, but unsupported (yet?) arguments.
    kArgUnsupported = frozenset((
        "-iprefix",
        "--include-prefix",
        "-iwithprefix",
        "-iwithprefixbefore",
        "--include-with-prefix",
        "--include-with-prefix-after",
        "--include-with-prefix-before",
        "-working-directory",
        "--system-header-prefix",
    ))
    kPfxArgUnsupported = (
        "-working-directory=",
        "-iprefix",
        "--include-prefix=",
        "-iwithprefix",
        "--include-with-prefix-after=",
        "--include-with-prefix=",
        "-iwithprefixbefore",
        "--include-with-prefix-before=",
        "-iwithsysroot",
        "--system-header-prefix=",
        "--no-system-header-prefix=",
    )

    # To properly find source files in compiler's args, we would have to implement a complete parser,
    # which is infeasible. We use an extension/suffix based heuristic and a flags blacklist instead.
    # We aren't interested in all sources, but only those contributing information visible from C++.
    # https://gcc.gnu.org/onlinedocs/gcc/Overall-Options.html
    kExtOfSource = (
        ".c",
        ".i",
        ".ii",
        ".m",
        ".mi",
        ".mm",
        ".M",
        ".mii",
        ".cc",
        ".cp",
        ".cxx",
        ".cpp",
        ".CPP",
        ".c++",
        ".C",
        # C++20 modules
        ".ixx",
        ".cppm",
        ".cxxm",
        ".c++m",
        ".ccm",
    )

    # if a path/file argument is preceded by any of the following flags, the arg is not a source
    kArgIsNotSource = frozenset(("-main-file-name",))

    # If any of these arguments present in a compiler invocation, then the whole invocation should
    # be ignored (information querying arguments)
    # isolated dash is for compilation from stdin - there's nothing we can do with it
    kArgToIgnoreInvocation = frozenset(("--version", "-"))
    # when there's only one arg and it matches anything below, ignore the invocation
    kSoleArgToIgnoreInvocation = frozenset(("-v", "--verbose"))
    # if an arg starts with these substring, the invocation will be ignored
    kArgStartsWithIgnoreInvocation = ("-print", "--print")
    # with the exception of the below args
    kArgStartsWithIgnoreInvocationException = ("--print-missing-file-dependencies",)

    # greedy match repeatedly blocks ending on escaped quote \" literal, or that doesn't contain
    # quotes at all until first unescaped quote
    @staticmethod
    def _makeRInQuotes(capture_inner: bool, no_begin_end: bool) -> str:
        not_word_backslash = r"(?<=\W)(?<!\\)"
        return (
            # either start of string, or NOT a word, or not a backslash
            (not_word_backslash if no_begin_end else r"(?:^|" + not_word_backslash + ")")
            + r"(?P<quote>['\"])"  # starting quote
            + ("(" if capture_inner else "(?:")  # start of capture/group
            # WARNING: never end a comment inside extended regexp with a backslash, even in raw
            # strings - it'll be treated as line continuation char
            + r"""
            (?:(?:(?!(?P=quote)).)*(?:\\(?P=quote))*)*  # any sequence of not quotes and escaped quotes
            (?<!\\)         # no escaping backslash in front of the ending quote
            )               # end of capture/group
            (?P=quote)      # ending quote
            """
            # after the quote either end of string, or NOT a word
            + (r"(?=\W)" if no_begin_end else r"(?:$|(?=\W))")
        )

    _r_in_quotes = re.compile(_makeRInQuotes(True, False), re.VERBOSE)
    # greedy match [] with any chars inside of ""
    _r_in_braces = re.compile(r"^\[(?:[, ]*" + _makeRInQuotes(False, True) + r")*\]", re.VERBOSE)
    # note that dependency of _r_in_braces on essentially _r_in_quotes definition is important as
    # it allows to produce a string on which we can safely match individual arguments with _r_in_quotes.findall()
    # in a batch, instead one by one to ensure correctness of parsing.

    """ from https://gcc.gnu.org/onlinedocs/gcc/Overall-Options.html:
    Options in file are separated by whitespace. A whitespace character may be included
    in an option by surrounding the entire option in either single or double quotes.
    Any character (including a backslash) may be included by prefixing the character to
    be included with a backslash. The file may itself contain additional @file options;
    any such options will be processed recursively."""
    # _r_options = re.compile(
    #    r"([^\s'\"]+|'(?:(?:[^']*\\'|[^']*)*)'|\"(?:(?:[^\"]*\\\"|[^\"]*)*)\")(?:\s+|$)"
    # )
    _r_options = re.compile(
        r"""\s*
        (?:
        ([^\s'\"]+) |
        (?:"""
        + _makeRInQuotes(True, True)
        + r""")
        )
        (?:\s+|$)""",
        re.VERBOSE,
    )

    # ending of execve() line
    _r_execve_end = re.compile(r"\)\s*=\s*0\s*$")

    @staticmethod
    def _leadingPlusToDash(s: str) -> str:
        if s.startswith("++"):
            return "--" + s[2:]
        elif s.startswith("+"):
            return "-" + s[1:]
        return s

    def __init__(self, Con: LoggingConsole, args: argparse.Namespace) -> None:
        self.Con = Con
        self._compilers = args.compiler
        self._do_other = args.other_commands
        self._cwd = os.path.realpath(args.cwd)
        self._test_files = not args.ignore_not_found
        self._discard_outputs_with_pfx = tuple(
            s for s in args.discard_outputs_with_pfx if len(s) > 0
        )
        self._discard_sources_with_pfx = tuple(
            s for s in args.discard_sources_with_pfx if len(s) > 0
        )
        self._discard_args_with_pfx = tuple(
            BaseParser._leadingPlusToDash(s) for s in args.discard_args_with_pfx if len(s) > 0
        )
        self._discard_args = self._makeDiscardArgs(Con, args.discard_args)
        self._do_dupes_check = args.enable_dupes_check

        if self._test_files and not os.path.isdir(self._cwd):
            Con.warning(
                "Compilation directory '",
                self._cwd,
                "' doesn't exist. If you used --cwd option, check its correctness. "
                "Resulting json will likely be invalid.",
            )

        self._running_pids: dict[int, ProcessProps] = {}

        self.compile_commands: list[CompileCommand] = []
        self.compile_cmd_time: list[float] = []
        self.other_commands: list[OtherCommand] = []
        self.other_cmd_time: list[float] = []
        # errors = {} # error_code -> array of line_idx where it happened

        self._parseLog(args.log_file)

    @staticmethod
    def _makeDiscardArgs(
        Con: LoggingConsole, discard_args: list[str] | None
    ) -> dict[str, str | None]:
        if not discard_args:
            return {}
        ret: dict[str, str | None] = {}

        def _upd(n, v=None):
            nonlocal ret
            if n in ret and ret[n] != v:
                Con.warning(
                    "in parsing --discard_args: argument",
                    n,
                    "is already specified with value",
                    ret[n],
                    ". The value is replaced with",
                    v,
                )
            ret[n] = v

        for da in discard_args:
            da = BaseParser._leadingPlusToDash(da)

            splt = da.split("|", maxsplit=1)
            if len(splt) == 1:
                if da.startswith("-D"):
                    _upd("--define-macro=" + da[2:])
                _upd(da)
            else:
                assert len(splt) == 2
                name, value = splt
                if name == "-D":
                    _upd("--define-macro", value)
                _upd(name, value)

        return ret

    def _parseLog(self, log_file: str) -> None:
        # match the start of the log string: (<pid>) (<time.stamp>) (execve|execveat|exited...)
        r_exec_or_exit = re.compile(
            r"^(?P<pid>\d+)\s+(?P<unix_ts>\d+)\.(?P<unix_ts_ms>\d+)\s+(?P<call>execve|execveat|\+\+\+ exited with (?P<exit_code>\d+) \+\+\+)"
        )

        # maps source->{output: (args_str, line_num)} to verify that commands are unique
        self._seen_compile: dict[str, dict[str | None, tuple[str, int]]] = {}
        self._seen_other: dict[str | None, tuple[str, int]] = {}  # just output->(args_str,line_num)

        self._unsupported_args = set()  # set of found unsupported args
        self._num_dropped_args_by_pfx = 0  # for a report on how many args by pfx were dropped
        self._num_dropped_args_by_match = 0

        with rich.progress.open(
            log_file, "r", description="Parsing strace log file...", console=self.Con
        ) as file:
            # sometimes strace breaks reporting of a single execve() call in two lines. All known
            # cases of that have the first line ending on `<unfinished ...>` literal and the next
            # line starting with `)` literal. We have to handle this
            unfinished_line: str | None = None
            unfinished_args: tuple | None = None
            for line_idx, line in enumerate(file):
                line = line.strip()  # dropping line-ends and other whitespaces
                # handling line continuations
                if unfinished_line is not None:
                    prev_line = unfinished_line
                    unfinished_line = None
                    if line.startswith(")"):
                        self._handleExec(*unfinished_args, prev_line + line)
                        continue
                    else:
                        self.Con.error(
                            "Line",
                            line_idx + 1,
                            "has unexpected continuation pattern. Treating this and prev lines as independent.",
                        )
                        self._handleExec(*unfinished_args, prev_line)
                        # falldown to processing below

                match_exec_or_exit = r_exec_or_exit.match(line)
                if not match_exec_or_exit:
                    unfinished_line = None
                    continue  # nothing to do here

                pid = int(match_exec_or_exit.group("pid"))
                ts = float(match_exec_or_exit.group("unix_ts")) + float(
                    1e-6 * int(match_exec_or_exit.group("unix_ts_ms"))
                )
                call = match_exec_or_exit.group("call")
                exit_code = match_exec_or_exit.group("exit_code")  # could be None

                if call.startswith("+++ "):
                    unfinished_line = None
                    if pid not in self._running_pids:
                        continue  # this must be not a process we care about
                    self._handleExit(pid, ts, exit_code, line_idx + 1)
                else:
                    # handle execve/execveat here
                    if line.endswith("<unfinished ...>"):
                        unfinished_line = line[match_exec_or_exit.end() :]
                        unfinished_args = (call, pid, ts, line_idx + 1)
                        self.Con.trace(
                            "Line",
                            line_idx + 1,
                            "pid",
                            pid,
                            "is unfinished. Deferring processing to the next line.",
                        )
                    else:
                        self._handleExec(
                            call, pid, ts, line_idx + 1, line[match_exec_or_exit.end() :]
                        )
            if unfinished_line is not None:
                self.Con.error(
                    "Previous line is marked as unfinished, but this was the last line. Trying to handle it"
                )
                self._handleExec(*unfinished_args, unfinished_line)

        # finishing unfinished processes
        for pid in list(self._running_pids.keys()):  # must rematerialize since exit() deletes them
            self._handleExit(pid, 0.0, None, 0)

        assert 0 == len(self._running_pids)
        n_cc = len(self.compile_commands)
        n_lc = len(self.other_commands)
        if n_cc == 0 and n_lc == 0:
            self.Con.print(
                "No compiler invocation were found in the log. If you're using a custom compiler, pass it in --compiler option."
            )
        else:
            self.Con.print(n_cc, "compilation commands found")
            if self._do_other:
                self.Con.print(n_lc, "other commands found")

        if self._num_dropped_args_by_pfx:
            self.Con.info(
                "In total ",
                self._num_dropped_args_by_pfx,
                "compiler arguments were removed according to the following --discard_args_with_pfx specification:",
                self._discard_args_with_pfx,
            )

        if self._num_dropped_args_by_match:
            self.Con.info(
                "In total ",
                self._num_dropped_args_by_match,
                "compiler arguments were removed according to the following --discard_args specification:",
                self._discard_args,
            )

        if self._unsupported_args:
            self.Con.warning(
                f"Found use of {len(self._unsupported_args)} unsupported arguments: {sorted(self._unsupported_args)}. "
                "These were ignored even though they might affect correctness of the resulting .json. "
                "If you think these should be supported, consider making a PR or reporting an issue."
            )

        # cleanup
        del self._num_dropped_args_by_match
        del self._num_dropped_args_by_pfx
        del self._unsupported_args
        del self._seen_other
        del self._seen_compile

    def _handleExit(self, pid: int, ts: float, exit_code: str | None, line_num: int) -> None:
        # negative exit code means the process termination was not found in the log
        (start_ts, start_line_num, is_other, cmd_idx, n_sources) = self._running_pids[pid]

        is_exit_logged = line_num > 0
        if is_exit_logged:  # <=0 line_idx is used when we didn't find the process exit in the log
            assert exit_code is not None, (
                f"Line {line_num}: pid {pid} exited without an exit code. This violates parser assumptions"
            )
            # if exit code isn't set, something is very wrong with the regexp or the log file,
            # so there's no point to try to continue. However, even if the exit code is non-zero,
            # we could at least save the other commands to compile_commands.json.

            if exit_code != "0":
                self.Con.warning(
                    f"Line {line_num}: pid {pid} (started at line {start_line_num}) exited with "
                    f"non-zero exit code {exit_code}. This might mean the build wasn't successful "
                    "and the resulting .json might be incomplete."
                )

            if ts < start_ts:
                # depending on used clock type, this might happen due to clock adjustments
                self.Con.warning(
                    f"Line {line_num}: pid {pid} (started at line {start_line_num}) exited at time "
                    f"{ts:.6f} which is before it started at "
                    f"{start_ts:.6f}. Continuing, but the log file might be malformed."
                )
                # todo: save this to errors
        else:
            self.Con.warning(
                f"pid {pid} (started at line {start_line_num}) didn't log its exit. "
                "This might mean the log file is incomplete and hence so is the resulting .json."
            )

        duration = ts - start_ts if is_exit_logged else 0.0
        if is_other:
            self.other_cmd_time[cmd_idx] = duration
        else:
            self.compile_cmd_time[cmd_idx : cmd_idx + n_sources] = [duration] * n_sources

        del self._running_pids[pid]

    def _testPathExists(self, arg: str, line_num: int, pid: int, args_str: str) -> None:
        if self._test_files and not unescapedPathExists(self._cwd, arg):
            self.Con.warning(
                f"Line {line_num}: pid {pid} uses argument '{arg}' "
                "which doesn't exist. This might mean the build system is misconfigured "
                "or the log file is incomplete and hence so is the resulting .json. "
                f"Full command args are: {args_str}"
            )

    def _handleExec(self, call: str, pid: int, ts: float, line_num: int, line: str) -> None:
        assert pid not in self._running_pids  # should be checked by the caller
        """assert call in ("execve", "execveat"), (
            f"Line {line_idx}: pid {pid} made call {call}. The code is inconsistent "
            "with rExecOrExit regexp"
        )"""
        assert call == "execve", (
            "execveat() handling is not implemented yet, consider making a PR or report "
            "an issue supplying a log file with execveat() calls"
        )
        assert line[0:1] == "(", "Unexpected format of the {call} syscall in the log file"

        # this search is very slow, but it can't be simplified and it's an important diagnostics feature.
        if not self._r_execve_end.search(line):
            self.Con.warning(f"Line {line_num}: pid {pid}: unexpected end of '{line}'.")

        # extract the first argument of execve, which is the executable path
        match_filepath = self._r_in_quotes.match(line[1:])
        assert match_filepath, (
            f"Line {line_num}: pid {pid} made call {call} but the argument '{line}' can't be parsed. "
            "The log file is malformed or _r_in_quotes regexp is incorrect"
        )

        # unescaping quotes and other symbols.
        compiler_path = unescapePath(match_filepath.group(2))
        if (
            compiler_path not in self._compilers.fullpaths
            and os.path.basename(compiler_path) not in self._compilers.basenames
        ):
            return  # not a compiler we care about

        # finding execv() args in the rest of the line
        args_start_pos = match_filepath.end() + 3
        # +1 since match_filepath is matched on args[1:]
        assert line[match_filepath.end() + 1 : args_start_pos + 1] == ", [", (
            f"Unexpected format of the {call} syscall in the log file"
        )
        # we can't simply search for the closing ] because there might be braces in file names and
        # they don't have to be shell-escaped
        match_args = self._r_in_braces.match(line[args_start_pos:])
        assert match_args, (
            f"Line {line_num}: pid {pid} made call {call} but the arguments array couldn't be parsed. "
            "The log file is malformed or _r_in_braces regexp is incorrect"
        )

        args_str = match_args.group()

        # Extracting args from the args_str. We can't simply split by ", " because there might be
        # such sequence in file names. So we use the same rInQuotes regexp to extract them one by one.
        # In a sense, it's a duplication of application of the same regexp as above, but we must
        # scope the search to the inside of the braces only
        args = re.findall(self._r_in_quotes, args_str)
        args = [inner for _, inner in args]

        if self._shouldIgnoreInvocation(args, line_num, pid, args_str):
            return
        args = self._expandAtFile(args, line_num, pid)

        # now walking over the args and checking existence of those that we know to be files or dirs.
        # Also getting arguments of some important options, if they are present
        next_is_path = False
        next_is_output = False
        arg_output = None
        sources: list[str] = []
        discard_arg_idx: list[int] = []

        skip_next = False
        for idx, arg in enumerate(args):
            if skip_next:
                skip_next = False
                continue
            elif next_is_path:  # no point to check anything if it's set
                next_is_path = False
                self._testPathExists(arg, line_num, pid, args_str)
                if next_is_output:
                    next_is_output = False
                    if arg_output is not None:
                        self.Con.warning(
                            f"Line {line_num}: pid {pid} uses multiple output options. "
                            f"This is unusual, taking the last one. Full command args are: {args_str}"
                        )
                    arg_output = arg  # it's already escaped
                continue
            elif self._discard_args_with_pfx and arg.startswith(self._discard_args_with_pfx):
                # argument blacklist is the first check
                self._num_dropped_args_by_pfx += 1
                discard_arg_idx.append(idx)
                continue
            elif arg in self._discard_args:
                discard_val = self._discard_args[arg]
                if discard_val is None:
                    self._num_dropped_args_by_match += 1
                    discard_arg_idx.append(idx)
                    continue
                else:
                    if idx + 1 < len(args) and discard_val == args[idx + 1]:
                        self._num_dropped_args_by_match += 2
                        discard_arg_idx.extend([idx, idx + 1])
                        skip_next = True
                        continue
                    # else fall below

            if arg in self.kArgIsPath and (
                arg not in self.kCheckArgForSysrootSpec or not arg.startswith(self.kSysrootSpec)
            ):  # do nothing for sysroot spec, as it's a subdir of tested -sysroot
                next_is_path = True
                if arg in self.kOutputArgs:
                    next_is_output = True
            elif m_pfx_arg := self.r_pfx_arg_is_path.match(arg):
                path_part = arg[m_pfx_arg.end() :]
                if path_part:
                    self._testPathExists(path_part, line_num, pid, args_str)
                    if "--output=" == m_pfx_arg.group(1):
                        if arg_output is not None:
                            self.Con.warning(
                                f"Line {line_num}: pid {pid} uses multiple output options. "
                                f"This is unusual, taking the last one. Full command args are: {args_str}"
                            )
                        arg_output = path_part  # it's already escaped
                else:
                    self.Con.error(
                        f"Line {line_num}: pid {pid} uses '{arg}' argument without a path. "
                        f"This is likely a bug. Full command args are: {args_str}"
                    )
            elif arg in self.kArgUnsupported or arg.startswith(self.kPfxArgUnsupported):
                self._unsupported_args.add(arg)
            elif (
                arg.endswith(self.kExtOfSource)
                and idx > 0
                and args[idx - 1] not in self.kArgIsNotSource
            ):
                sources.append(arg)

        if discard_arg_idx:
            for i in reversed(discard_arg_idx):
                del args[i]

        n_sources = len(sources)
        is_other = n_sources == 0

        if not self._do_other and is_other:
            return  # not interested in other commands

        if arg_output is None:
            if self._do_dupes_check:
                # arg_output could be not specified at all in which case there're rules for making a default
                # value for it, The problem is, the rules are based on input files and we can reliably parse
                # only a subset of possible inputs.
                self.Con.warning(
                    f"Line {line_num} pid {pid}: output wasn't explicitly set for a call '{args_str}'."
                    "\nOutput based duplicate checks might yield false positives."
                )
        else:
            if self._discard_outputs_with_pfx and arg_output.startswith(
                self._discard_outputs_with_pfx
            ):
                self.Con.trace(
                    f"Line {line_num} pid {pid}: call '{args_str}' is ignored due to args.discard_outputs_with_pfx"
                )
                return

        if n_sources > 1:
            self.Con.warning(
                f"Line {line_num} pid {pid}: call '{args_str}' has {n_sources} sources"
            )

        if self._discard_sources_with_pfx:
            sources = [s for s in sources if not s.startswith(self._discard_sources_with_pfx)]
            if not sources:
                self.Con.trace(
                    f"Line {line_num} pid {pid}: call '{args_str}' is ignored due to args.discard_sources_with_pfx"
                )
                return

        # TODO: do we need to fix the first argument in args to be the same as the one used in
        # execve()? It might be different depending how execve() was called.
        if is_other:
            if self._do_dupes_check:
                self._checkSameOther(args_str, line_num, arg_output)
            self.other_commands.append(OtherCommand(args, arg_output, line_num))
            cmd_idx = len(self.other_cmd_time)
            self.other_cmd_time.append(0.0)
        else:
            cmd_idx = len(self.compile_cmd_time)
            for src in sources:
                if self._do_dupes_check:
                    self._checkSameCompile(args_str, line_num, arg_output, src)
                self.compile_commands.append(CompileCommand(args, arg_output, src, line_num))
                self.compile_cmd_time.append(0.0)

        self._running_pids[pid] = ProcessProps(ts, line_num, is_other, cmd_idx, n_sources)

    def _shouldIgnoreInvocation(
        self, args: list[str], line_num: int, pid: int, args_str: str
    ) -> bool:
        args_len = len(args)
        if args_len <= 1:
            self.Con.error(
                f"Line{line_num} pid{pid}: invocation of a compiler '{args_str}' doesn't have arguments. Ignoring it",
            )
            return True

        if args_len == 2 and args[1] in self.kSoleArgToIgnoreInvocation:
            self.Con.trace(
                f"Line{line_num} pid{pid}: invocation '{args_str}' is ignored due to a sole arg in kSoleArgToIgnoreInvocation"
            )
            return True

        if any(1 for a in args if a in self.kArgToIgnoreInvocation):
            self.Con.trace(
                f"Line{line_num} pid{pid}: invocation '{args_str}' is ignored due to an arg in kArgToIgnoreInvocation"
            )
            return True

        if any(
            1
            for a in args
            if a.startswith(self.kArgStartsWithIgnoreInvocation)
            and a not in self.kArgStartsWithIgnoreInvocationException
        ):
            self.Con.trace(
                f"Line{line_num} pid{pid}: invocation '{args_str}' is ignored due to an arg satisfying kArgStartsWithIgnoreInvocation"
            )
            return True

        return False

    def _expandAtFile(self, args: list[str], line_num: int, pid: int) -> list[str]:
        at_idxs = [i for i, s in enumerate(args) if s.startswith("@")]
        added = 0
        for i in at_idxs:
            fname = toAbsPathUnescape(self._cwd, args[i][1:])
            if os.path.isfile(fname):
                fsize = os.path.getsize(fname)
                if fsize > 64 * 1024:  # randomly sufficient threshold
                    self.Con.info(
                        f"Line {line_num}: pid {pid} has @file argument#{i} '{args[i]}' that "
                        "points to a file of size",
                        fsize,
                        ". That seems a bit too much, perhaps something is wrong?",
                    )
                with open(fname, "r") as file:
                    file_content = file.read()
                # self.Con.debug("@file", args[i], "  -->  ", file_content)
                newargs = []
                ofs = 0
                # have to sequentially match one after the other to ensure everything is parsed as
                # expected. Doing .findall() will just skip parts that don't match and this would
                # miss bugs
                m = self._r_options.match(file_content)
                while m:
                    noq, inq = m.group(1, 3)
                    assert int(noq is None) + int(inq is None) == 1
                    newargs.append(noq if noq is not None else inq)
                    ofs += m.end()
                    m = self._r_options.match(file_content[ofs:])
                if len(file_content) != ofs:
                    self.Con.error(
                        f"Line {line_num}: pid {pid} has @file argument#{i} '{args[i]}' "
                        "parsing of which ended prematurely."
                    )
                # self.Con.debug(newargs)
                newargs = self._expandAtFile(newargs, line_num, pid)  # recursive expansion
                # ^^ might produce puzzling messages, but that's a TODO for later

                ni = added + i
                added += len(newargs) - 1
                args[ni:ni] = newargs
                del args[added + i + 1]
                # self.Con.debug(args)

            else:
                self.Con.error(
                    f"Line {line_num}: pid {pid} has @file argument#{i} '{args[i]}' that doesn't "
                    "reference existing file. Processing might yield incomplete results."
                )
                # do nothing
        return args

    def _checkSameCompile(
        self, arg_str: str, line_num: int, arg_output: str | None, arg_compile: str
    ) -> bool:
        if arg_compile in self._seen_compile:
            outp = self._seen_compile[arg_compile]
            if arg_output in outp:
                prev_args, prev_line = outp[arg_output]
                output_repr = "<<not_determined>>" if arg_output is None else arg_output
                if arg_str == prev_args:
                    self.Con.warning(
                        "For source '",
                        arg_compile,
                        "' the same output '",
                        output_repr,
                        "' is produced by the second instance of the ~same compilation command vvv\n",
                        arg_str,
                        f"\n^^^. First line was @line#{prev_line}, now it's line#{line_num}. "
                        "This might be benign, but this isn't normal for a correct build system. ",
                    )
                    return True
                else:
                    self.Con.error(
                        "For source '",
                        arg_compile,
                        "' the same output '",
                        output_repr,
                        f"' is generated by different compilation commands: the first recorded was @line#{prev_line} vvv\n",
                        prev_args,
                        f"\n^^^ and now @line#{line_num} it's vvv\n",
                        arg_str,
                        "\n^^^. This isn't normal and could mean that several build systems were "
                        "executed.",
                    )
            else:
                outp[arg_output] = (arg_str, line_num)
        else:
            self._seen_compile[arg_compile] = {arg_output: (arg_str, line_num)}
        return False

    def _checkSameOther(self, arg_str: str, line_num: int, arg_output: str | None) -> bool:
        if arg_output in self._seen_other:
            prev_args, prev_line = self._seen_other[arg_output]
            output_repr = "<<not_determined>>" if arg_output is None else arg_output
            if arg_str == prev_args:
                self.Con.warning(
                    "The same output '",
                    output_repr,
                    "' is produced by the second instance of the same other command vvv\n",
                    arg_str,
                    f"\n^^^. First was @line#{prev_line}, now it's line#{line_num}. "
                    "This might be benign, but this isn't normal for a correct build system. ",
                )
                return True
            else:
                self.Con.error(
                    "The same output '",
                    output_repr,
                    f"' is generated by different other commands: the first recorded was @line#{prev_line} vvv\n",
                    prev_args,
                    f"\n^^^ and now @line#{line_num} it's vvv\n",
                    arg_str,
                    "\n^^^. This isn't normal and could mean that several build systems were "
                    "executed.",
                )
        else:
            self._seen_other[arg_output] = (arg_str, line_num)
        return False

    def storeJsons(
        self, dest_dir: str, save_duration: bool, save_line_num: bool, sfx: str = ""
    ) -> None:
        if not os.path.isdir(dest_dir):
            raise YacceException(f"Destination directory '{dest_dir}' doesn't exist")

        storeJson(
            self.Con,
            dest_dir,
            True,
            self.compile_commands,
            self.compile_cmd_time if save_duration else None,
            self._cwd,
            save_line_num,
            file_sfx=sfx,
        )
        if self._do_other:
            storeJson(
                self.Con,
                dest_dir,
                False,
                self.other_commands,
                self.other_cmd_time if save_duration else None,
                self._cwd,
                save_line_num,
                file_sfx=sfx,
            )


def storeJson(
    Con: LoggingConsole,
    path: str,
    is_compile_commands: bool,
    commands: list[CompileCommand] | list[OtherCommand],
    cmd_times: list[float] | None,
    cwd: str,
    save_line_num: bool,
    file_sfx="",
):
    filename = os.path.join(
        path, ("compile" if is_compile_commands else "other") + f"_commands{file_sfx}.json"
    )
    if os.path.exists(filename):
        os.remove(filename)

    if not commands:
        assert not cmd_times
        Con.info("storeJson() got empty list to save into '", filename, "'")
        return

    save_duration = cmd_times is not None
    assert not save_duration or len(commands) == len(cmd_times)

    e = next(iter(commands))
    is_other = isinstance(e, OtherCommand)
    assert is_other or isinstance(e, CompileCommand)
    assert int(is_other) + int(is_compile_commands) == 1

    cwd = cwd.replace('"', '\\"')
    with open(filename, "w") as f:
        f.write("[\n")
        for idx, cmd_tuple in enumerate(commands):
            f.write(("," if idx > 0 else "") + "{\n")
            f.write(f' "directory": "{cwd}",\n')
            if is_other:
                args, arg_output, line_num = cmd_tuple
            else:
                args, arg_output, arg_compile, line_num = cmd_tuple
                f.write(f' "file": "{arg_compile}",\n')

            if save_line_num:
                f.write(f' "line_num": {line_num},\n')
            if save_duration:
                f.write(f' "duration_s": {cmd_times[idx]:.6f},\n')
            if arg_output is not None:
                f.write(f' "output": "{arg_output}",\n')

            args_str = '", "'.join(args)
            f.write(f' "arguments": ["{args_str}"]\n')
            f.write("}\n")

        f.write("]\n")
    Con.print("Written", len(commands), "commands to", filename)


def unescapePath(path: str) -> str:
    # not sure this is correct
    return path.encode("latin1").decode("unicode_escape")


def escapePath(path: str) -> str:
    return path.encode("unicode_escape").decode("latin1")


def toAbsPathUnescape(cwd: str, path: str) -> str:
    path = unescapePath(path)
    if not os.path.isabs(path):
        path = os.path.join(cwd, path)
    return os.path.realpath(path)  # resolve symlinks so isfile() or isdir() works properly


def unescapedPathExists(cwd: str, path: str) -> bool:
    return os.path.exists(toAbsPathUnescape(cwd, path))



import argparse
import sys

from yacce import common
from yacce.mod_bazel import mode_bazel
from yacce.mod_from_log import mode_from_log


def _getModeHelpString(mode: str) -> str:
    return f"Hint: use 'yacce {mode} --help' to get CLI arguments help."


# default mode is the first
kModes = {
    "bazel": "Runs a given build system based on Bazel in a shell and extracts compile_commands.json "
    "from it (possibly with individual compile_commands.json for each external dependency).\nThis is "
    "a default mode activated if the mode specification is just omitted.\n"
    + _getModeHelpString("bazel"),
    "from_log": "[dbg!] Generates a possibly NON-WORKING(!) compile_commands.json from a strace log file.\n"
    "This mode features the most generic way to parse strace output and since the log generally "
    "lacks some important information (such as the working directory in case of a Bazel), it may "
    "produce a non-working compile_commands.json. The mode is primarily intended for debugging "
    "purposes as it doesn't use any knowledge about the build system used and just parses the strace "
    "log file and turns it into compile_commands.json as is.\n" + _getModeHelpString("from_log"),
}

kModeFuncs = {"bazel": mode_bazel, "from_log": mode_from_log}


def getModeArgs():
    """Makes mode + some early options parser.
    Possible modes are:
    - 'from_log': just takes strace log file and tries to make a compile_commands.json from it.
    - 'bazel': takes a bunch of options and runs a build system passed after -- argument assuming
        it's based on Bazel. This is a default mode activated if the first script argument doesn't
        match to any of the defined modes.
    - <add here other build systems when needed as separate modes>.
    - 'help | --help | -h': prints help message and exits.
    """
    # unfortunately, for some unexplained and dumb reason, argparse doesn't support parsing known
    # arguments only up to the first unknown argument, so we have to manually control for that.
    parser = argparse.ArgumentParser(
        prog="yacce",
        description=common.kMainDescription,
        formatter_class=common.BetterHelpFormatter,
        #argparse.RawTextHelpFormatter, #RawDescriptionHelpFormatter,
    )

    # dumb, but argparse doesn't have a way to query for defined flags
    definedFlags = {"--debug"}
    parser.add_argument(
        "--debug",
        help="Minimum debug level to show. 0 is the most verbose.\n"
        f"Default level is (info=) %(default)s. Setting it higher than a (warning=) {common.LoggingConsole.LogLevel.Warning.value} is not recommended.",
        type=int,
        choices=range(0, common.LoggingConsole.LogLevel.Critical.value + 1),
        default=common.LoggingConsole.LogLevel.Info.value,
    )

    definedFlags |= {"--colors", "--no-colors"}
    parser.add_argument(
        "--colors",
        help="Controls if the output could be colored.",
        action=argparse.BooleanOptionalAction,
        default=True,
    )

    modes = parser.add_subparsers(
        help='Modes of operation. Use "--help" with each mode to get more information.'
    )

    for mode, description in kModes.items():
        p = modes.add_parser(mode, help=description)
        p.add_argument("--mode", dest="mode", default=mode, help=argparse.SUPPRESS)

    if len(sys.argv) <= 2:  # there's always more than 1 arg
        parser.print_help()
        sys.exit(2)

    not_found = 1
    for first_rest, arg in enumerate(sys.argv[1:]):
        if arg not in definedFlags and arg not in kModes and not arg.isdigit():
            not_found = 0
            break
    first_rest += 1 + not_found
    if first_rest >= len(sys.argv):  # mode specific args always exist
        parser.print_help()
        sys.exit(2)

    return parser.parse_args(sys.argv[1:first_rest]), sys.argv[first_rest:]


def main():
    args, unparsed_args = getModeArgs()
    Con = common.LoggingConsole(
        no_color=not args.colors, log_level=common.LoggingConsole.LogLevel(args.debug)
    )
    Con.yacce_begin()

    Con.debug("mode args:", args)
    Con.debug("args past the mode:", unparsed_args)

    if not hasattr(args, "mode"):
        Con.debug("Mode is not specified, using the default")
    mode = getattr(args, "mode", next(iter(kModes)))

    try:
        ret = kModeFuncs[mode](Con, args, unparsed_args)
        common.warnClangdIncompatibilitiesIfAny(Con, args)

        Con.debug(f"Exiting with code {ret}")
        sys.exit(ret)
    except common.YacceException as e:
        Con.critical(repr(e))
        Con.debug("Exiting with code 1")
        sys.exit(1)


if __name__ == "__main__":
    main()



import argparse
import os
import re

from .common import (
    addCommonCliArgs,
    BaseParser,
    BetterHelpFormatter,
    kMainDescription,
    LoggingConsole,
    makeCompilersSet,
    YacceException,
)


def _fixCwdArg(Con: LoggingConsole, args: argparse.Namespace) -> argparse.Namespace:
    """Fixes the --cwd argument if it's relative path spec.
    If --cwd is not set, returns the directory of the log file.
    Also tests existence of the directory if it is set and not ignored, and modifies args.ignore_not_found
    if the directory doesn't exist.
    """
    assert isinstance(args, argparse.Namespace) and hasattr(args, "ignore_not_found")
    assert hasattr(args, "log_file") and isinstance(args.log_file, str)

    if hasattr(args, "cwd") and args.cwd:
        cwd = (
            args.cwd if os.path.isabs(args.cwd) else os.path.dirname(args.log_file) + "/" + args.cwd
        )
    else:
        cwd = os.path.dirname(args.log_file)

    cwd = os.path.realpath(cwd)
    if not args.ignore_not_found and not os.path.isdir(cwd):
        Con.warning(
            f"Working directory '{cwd}' does not exist, will not check file existence. "
            "Resulting compile_commands.json will likely be incorrect."
        )
        setattr(args, "ignore_not_found", True)

    setattr(args, "cwd", cwd)
    return args


def _getArgs(
    Con: LoggingConsole, args: argparse.Namespace, unparsed_args: list
) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="yacce from_log",
        description=kMainDescription
        + "\n\nMode 'from_log' is a supplementary mode that generates a compile_commands.json from "
        "a strace log file without using any additional information about build system.\n"
        "ATTENTION: this mode is intended for debugging purposes only and most likely will not "
        "produce a correct compile_commands.json due to a lack of information about the build process details.\n"
        "If you want to regenerate compile_commands from a log file for Bazel, use 'yacce bazel --from_log' instead.",
        formatter_class=BetterHelpFormatter,
        #argparse.RawTextHelpFormatter, #RawDescriptionHelpFormatter,
    )
    parser.add_argument("log_file", help="Path to the strace log file to parse.", type=str)
    parser = addCommonCliArgs(
        parser,
        {
            "cwd": "In the 'from_log' mode a relative path specification is resolved to "
            "the absolute path using a directory of the log file.\n"
            "Default: directory of the log file.",
            "dest_dir":" Default: current working directory."
        },
    )
    args = parser.parse_args(unparsed_args, namespace=args)

    if args.log_file is None or not os.path.isfile(args.log_file):
        raise YacceException(f"Log file '{args.log_file}' is not specified or does not exist.")
    
    if not args.dest_dir:
        args.dest_dir = os.getcwd()

    args = _fixCwdArg(Con, args)
    setattr(args, "compiler", makeCompilersSet(args.compiler))
    return args


def mode_from_log(Con: LoggingConsole, args: argparse.Namespace, unparsed_args: list) -> int:
    args = _getArgs(Con, args, unparsed_args)

    p = BaseParser(
        Con, args.log_file, args.cwd, not args.ignore_not_found, args.compiler, args.other_commands
    )

    p.storeJsons(args.dest_dir, args.save_duration, args.save_line_num)    
    return 0



import argparse
import itertools
import os
import re
from rich.progress import Progress
import shutil
import signal
import subprocess
import sys

if "linux" == sys.platform:
    import pty
    import shlex

from .common import (
    addCommonCliArgs,
    BaseParser,
    BetterHelpFormatter,
    CompileCommand,
    escapePath,
    # OtherCommand,
    kMainDescription,
    LoggingConsole,
    makeCompilersSet,
    storeJson,
    unescapePath,
    YacceException,
)


def _getArgs(
    Con: LoggingConsole, args: argparse.Namespace, unparsed_args: list
) -> tuple[argparse.Namespace, list]:
    parser = argparse.ArgumentParser(
        prog="yacce bazel",
        description=kMainDescription
        + "\n\nMode 'bazel' is intended to generate compile_commands.json from tracing execution of "
        "a 'bazel build' or any other shell command invoking Bazel using the Linux's strace utility. Hence it "
        "only supports compilation of a single Bazel workspace (including its all external dependencies) "
        "happening locally. If you are using Bazel's remote "
        "caching feature, including '--disk_cache', please make sure you're starting with a clean "
        "cache, otherwise yacce won't see compilation of cache hits.",
        usage="yacce [global options] [bazel] [options (see below)] [-- shell command eventually invoking Bazel]",
        formatter_class=BetterHelpFormatter,
        # argparse.RawTextHelpFormatter, #RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "--log_file",
        help="Write strace log to and/or read it from this file.\nSee also '--from_log'.\n"
        "Default: 'strace.txt' in the current directory, %(default)s",
        metavar="path/to/file",
        default=os.path.join(os.getcwd(), "strace.txt"),
    )

    p_log = parser.add_argument_group(
        "Log mode, uses existing strace log and is mutually exclusive with the live mode"
    )
    excl1 = {"from_log"}
    p_log.add_argument(
        "--from_log",
        help="Toggles a mode in which yacce will only parse an existing log file specified by '--log_file', but "
        "will not invoke a build system to spy on.\n"
        "Mutually exclusive with '--keep_log' and requires no build system arguments passed (no "
        "'--' argument and anything after it).\n"
        "Default: not set, i.e. the mode is not activated.",
        action="store_true",
    )

    p_live = parser.add_argument_group(
        "Live mode (default), runs a Bazel build system and is mutually exclusive with the log mode"
    )
    excl2 = {"keep_log"}
    p_live.add_argument(
        "--keep_log",
        choices=["if_errors", "always", "never"],
        help="Determines conditions of keeping of the strace log file after yacce "
        "finishes. Mutually exclusive with '--from_log'.\n"
        "Default is 'always' as it might be useful to run yacce in the log mode with different "
        "arguments later on the same log file.",
    )
    excl2 |= {"clean"}
    p_live.add_argument(
        "--clean",
        choices=["always", "expunge", "never"],
        help="Determines, if a 'bazel clean' or 'bazel clean --expunge' commands should be executed "
        "before running the build.\n"
        "Note that if cleaning is disabled, cached (already compiled) sources will be invisible to "
        "yacce and hence will not make it into resulting compiler_commands.json! (iterative updates "
        "aren't supported yet)\n"
        "Default: not specified, yacce will ask if running 'bazel clean' is ok.",
    )

    parser.add_argument(
        "--external",
        choices=["ignore", "combine-with-overridden", "to-files", "to-external", "combine-all"],
        default="combine-with-overridden",
        help="Determines what to do when a compilation of a project's dependency source file (from 'external/' "
        "subdirectory) is found.\n"
        "- One option is to just to 'ignore' (remove) it and to leave in the resulting "
        "compile_commands.json *only* commands directly related to the project.\n"
        "- The default option 'combine-with-overridden' produces a single compile_commands.json containing "
        "main project's files as well as dependencies that are stored *outside* of their "
        "expected location at '$(bazel info output_base)/external/<repo>' (this typically happens when you "
        "override a dependency repo location for Bazel when you work on the project and its dependency simultaneously).\n"
        "- Option 'to-files' produces individual files nearby the main compile_commands.json, named like"
        "'compile_commands_ext_<repo>.json' for each external dependency '<repo>' (this might be useful "
        "for manual inspection).\n"
        "- Option 'to-external' differs from 'to-files' only in the location and naming of the resulting files. "
        "'to-external' produces an individual compile_commands.json in each external dependency's "
        "directory and is useful when you're going to open the dependency directory in a parallel IDE for a close inspection.\n"
        "- The 'combine-all' option just writes all compilation commands (for the main project and its "
        "dependencies) into a single file.\n"
        "See '--dest_dir' argument for a default location and/or override for the main project's compile_commands.json file.\n"
        "NOTE that since currently yacce doesn't properly process compiler invocations that aren't related "
        "to compiling C or C++ sources (such as linking only, or compiling ASM files), if '--other_commands' "
        "flag is specified, a compound other_commands.json (containing all other invocations of a "
        "compiler for the main project and its externals) will be saved nearby the main compile_commands.json "
        "irrespective of a value of this flag.",
    )

    parser.add_argument(
        "--bazel_command",
        default="bazel",
        help="Override which command to run to communicate with the instance of a Bazel for the build system.\n"
        "You don't typically need this argument, if you have bazelisk installed.\n"
        "To set the workspace directory see '--bazel_workspace' argument.\n"
        "Default: %(default)s",
        metavar='command_or_filepath'
    )

    parser.add_argument(
        "--bazel_workspace",
        help="Overrides Bazel workspace directory to set a current directory context for the bazel command (see "
        "'--bazel_command').\n" \
        "This is useful if yacce needs to be run from an outside of that "
        "workspace. Note that any dir under a real workspace would also work here.\n"
        "Default: a current working directory '%(default)s'.",
        metavar="path/to/dir",
        default=os.getcwd(),
    )

    parser.add_argument(
        "--build_cwd",
        help="By default, a shell command to start the build is invoked from the Bazel workspace "
        "directory (see '--bazel_workspace'). This argument allows to override that and set a "
        "different directory as a cwd for the build command.\n" \
        "Note that this is different from '--cwd' argument, which "
        "for the 'bazel --from_log' mode of yacce specifies a value of '$(bazel info execution_root)' directory.",
        metavar="path/to/dir",
    )

    parser.add_argument(
        "--build_shell",
        default="bash",
        help="Build command is executed by passing it to a shell. By default, '%(default)s' is "
        "used, but you can override that with this argument.",
        metavar="shell_to_use",
    )

    parser.add_argument(
        "--ensure_build_succeeds",
        help="By default yacce only warns if the build command fails (exits with a non-zero code) "
        "and it tries to process the strace log file to produce some results anyway. If you want to "
        "make sure that yacce will only use a full log of a successful build, set this argument to "
        "enforce yacce failure if the build fails.\n",
        action=argparse.BooleanOptionalAction,
        default=False,
    )

    parser = addCommonCliArgs(
        parser,
        {
            "cwd": "In the 'yacce bazel --from_log' mode, this argument overrides an output of "
            "'$(bazel info execution_root)' (i.e. this enables parsing of an existing log file without "
            "querying its build system). In the default live mode (when no '--from_log' argument is "
            "specified) this argument is either has to be unset, or match the output of '$(bazel info execution_root)'.",
            "dest_dir": " Default: directory of the log file (see '--log_file')",
            "other_commands": " Note that yacce currently does not implement attribution of other "
            "compilation commands to the project's external dependencies. I.e. all other commands related to "
            "compiling non-C++ sources and linking will be combined into a single other_commands.json "
            "file irrespective of '--external' argument setting.",
        },
    )

    # looking for -- in unparsed_args to save build system invocation args.
    if len(unparsed_args) < 2:  # the shortest is "-- build_script.sh"
        parser.print_help()
        sys.exit(2)

    not_found = 1
    for first_rest, arg in enumerate(unparsed_args):  # .index() with exception is a crap.
        if "--" == arg:
            not_found = 0
            break

    first_rest += 1 + not_found
    if first_rest < len(unparsed_args):
        mode_args = unparsed_args[: first_rest - 1]
        unparsed_args = unparsed_args[first_rest:]
    else:
        mode_args = unparsed_args
        unparsed_args = []

    args = parser.parse_args(mode_args, namespace=args)

    # checking mutually exclusive options
    if any(getattr(args, a, False) for a in excl1) and any(getattr(args, a, False) for a in excl2):
        parser.print_help()
        Con.critical("Options from these two lists are mutually exclusive: ", excl1, excl2)
        sys.exit(2)
    # taking care of defaults that weren't set due to mutual exclusion check. argparse is a crap too
    if args.keep_log is None:
        setattr(args, "keep_log", "always")
    # leaving .clean untouched so we could later ask the user
    # if args.clean is None:
    #    setattr(args, "clean", "always")

    setattr(args, "compiler", makeCompilersSet(args.compiler))
    return args, unparsed_args


class BazelParser(BaseParser):
    def __init__(self, Con: LoggingConsole, args: argparse.Namespace) -> None:
        Con.trace("Running base parser")
        do_test_files = not args.ignore_not_found
        setattr(args, "ignore_not_found", True)
        super().__init__(Con, args)

        setattr(args, "ignore_not_found", do_test_files)
        self._test_files = do_test_files

        Con.trace("Starting Bazel-specific processing...")
        self._update()

    def _makeFullPath(
        self, is_external: bool, subdir: str, path_ending: str, is_file: bool, warn=True
    ) -> tuple[str, bool]:
        exists = False
        fullpath = os.path.realpath(os.path.join(self._cwd, subdir, path_ending))
        if self._test_files:
            exists = (is_file and os.path.isfile(fullpath)) or (
                (not is_file) and os.path.exists(fullpath)
            )
            if is_external and not exists:
                fullpath2 = os.path.realpath(os.path.join(self._cwd, "../..", subdir, path_ending))
                exists = os.path.exists(fullpath2)
                if exists:
                    fullpath = fullpath2
                else:
                    if warn:
                        self.Con.warning(
                            f"External '{subdir}/{path_ending}' not found in both expected paths",
                            (fullpath, fullpath2),
                        )
        return fullpath, exists

    def _update(self) -> None:
        ext_paths: dict[str, str] = {}  # external canonical_name -> realpath
        extinc_paths: dict[str, str] = {}  # external include paths
        ext_ccs: dict[str, list[CompileCommand]] = {}
        ext_cctimes: dict[str, list[float]] = {}
        # TODO other commands!

        new_ccs: list[CompileCommand] = []  # new compile_commands for the project only
        new_ccs_time: list[float] = []

        notfound_inc: set[str] = set()

        r_external = re.compile(r"^(?:\.\/)?external\/")
        # generated files such as 'bazel-out/k8-opt/bin/external/<repo>/..' are also externals!
        # matches repo part in any external path spec. Not sure leading optional ./ is useful,
        # haven't seen it, but leaving it just in case
        r_any_external = re.compile(
            r"^(?:\.\/)?(?:bazel-[^\/]+\/[^\/]+\/bin\/)?external\/([^\/]+)\/"
        )
        # matches a whole external/... part in bazel-../../external/.. path spec
        r_bazel_external = re.compile(r"^(?:\.\/)?bazel-[^\/]+\/[^\/]+\/bin\/(external\/.+)$")

        def _fix_path(arg: str, argidx: int, args: list[str] | None) -> str:
            nonlocal extinc_paths, notfound_inc
            m_ext = r_any_external.match(arg)
            if m_ext:
                r = m_ext.group(1)
                if r not in extinc_paths:
                    repo_path, exists = self._makeFullPath(True, "external", r, False)
                    if self._test_files and not exists:
                        self.Con.warning(
                            f"External include repo '{r}' not found at expected path '{repo_path}'"
                        )
                    extinc_paths[r] = repo_path

            path, exists = self._makeFullPath(
                bool(r_external.match(arg)), "", unescapePath(arg), False, warn=False
            )
            if self._test_files and not exists:
                err = True
                if args is not None:
                    # ignoring existence test failure for same qualified args starting with bazel-out/k8-opt/bin/external/... dirs
                    # that exist as just normally qualified external/... args. This seems to be a bazel quirk
                    m_bzl_ext = r_bazel_external.match(arg)
                    if m_bzl_ext:
                        ext = m_bzl_ext.group(1)
                        qual = args[argidx - 1]  # can't be negative
                        # TODO: O(n^2), but maybe will improve later
                        for ai, a in enumerate(args[1:]):
                            # ai refs previous args element
                            if a == ext and qual == args[ai]:
                                err = False
                                break
                if err:
                    notfound_inc.add(arg)
            return escapePath(path)

        with Progress(console=self.Con) as progress:  # transient=True,
            task = progress.add_task(
                "Applying Bazel-specific transformations to the log...",
                total=len(self.compile_commands),
            )
            for ccidx, cc in enumerate(self.compile_commands):
                cctime = self.compile_cmd_time[ccidx]
                args, output, source, line_num = cc

                # deciding if this is external
                m_external = r_any_external.match(source)
                if m_external:
                    repo = m_external.group(1)
                    if repo not in ext_paths:
                        repo_path, exists = self._makeFullPath(True, "external", repo, False)
                        if self._test_files and not exists:
                            self.Con.warning(
                                f"External repo '{repo}' not found at expected path '{repo_path}'"
                            )
                        ext_paths[repo] = repo_path
                else:
                    repo = None

                # checking and updating the source path
                path, exists = self._makeFullPath(
                    bool(r_external.match(source)), "", unescapePath(source), True
                )
                if self._test_files and not exists:
                    self.Con.warning("Source file", path, "not found!")
                source = escapePath(path)
                # no need to check and update output

                # new_args = []
                next_is_path = False
                for argidx, arg in enumerate(args):
                    # resolving symlinks to reduce dependency on bazel's internal workspace structure
                    if next_is_path:
                        next_is_path = False
                        args[argidx] = _fix_path(arg, argidx, args)
                    elif arg in self.kArgIsPath and (
                        arg not in self.kCheckArgForSysrootSpec
                        or not arg.startswith(self.kSysrootSpec)
                    ):
                        next_is_path = True
                    elif m_pfx_arg := self.r_pfx_arg_is_path.match(arg):
                        path_part = arg[m_pfx_arg.end() :]
                        if path_part:
                            args[argidx] = m_pfx_arg.group() + _fix_path(path_part, argidx, None)
                    # new_args.append(arg)

                # new_cc = CompileCommand(new_args, output, source, line_num)
                new_cc = CompileCommand(args, output, source, line_num)
                if m_external:
                    ext_ccs.setdefault(repo, []).append(new_cc)
                    ext_cctimes.setdefault(repo, []).append(cctime)
                else:
                    new_ccs.append(new_cc)
                    new_ccs_time.append(cctime)
                progress.advance(task)

        if self.Con.will_log(self.Con.LogLevel.Debug):
            self.Con.debug(
                "Compiled dependencies list has",
                len(ext_paths),
                "entries:",
                {k: ext_paths[k] for k in sorted(ext_paths.keys())},
            )
            self.Con.debug(
                "Include dependencies list has",
                len(extinc_paths),
                "entries:",
                {k: extinc_paths[k] for k in sorted(extinc_paths.keys())},
            )

        ext_paths |= extinc_paths
        self.Con.print(
            "External dependencies list has",
            len(ext_paths),
            "entries:",
            {k: ext_paths[k] for k in sorted(ext_paths.keys())},
        )

        if len(notfound_inc) > 0:
            self.Con.warning(
                "These",
                len(notfound_inc),
                "paths are used in compiler includes, but doesn't exist. This might mean the "
                "build system is misconfigured, or the log file is incomplete, but sometimes it "
                "just happens and it's fine.",
                [v for v in sorted(notfound_inc)],
            )

        self._ext_paths = ext_paths
        self._ext_ccs = ext_ccs
        self._ext_cctimes = ext_cctimes
        self._new_cc = new_ccs
        self._new_cc_time = new_ccs_time

        # merging processed list back into the base class list storage
        self.compile_commands = list(itertools.chain(new_ccs, *ext_ccs.values()))
        self.compile_cmd_time = list(itertools.chain(new_ccs_time, *ext_cctimes.values()))

        # TODO other commands!

    def storeJsons(self, dest_dir: str, external: str, save_duration: bool, save_line_num: bool):
        # saving other_commands no matter what if requested
        if self._do_other:
            storeJson(
                self.Con,
                dest_dir,
                False,
                self.other_commands,
                self.other_cmd_time if save_duration else None,
                self._cwd,
                save_line_num,
            )

        if "combine-all" == external:
            storeJson(
                self.Con,
                dest_dir,
                True,
                self.compile_commands,
                self.compile_cmd_time if save_duration else None,
                self._cwd,
                save_line_num,
            )
            return

        if external in ("ignore", "to-files", "to-external"):
            storeJson(
                self.Con,
                dest_dir,
                True,
                self._new_cc,
                self._new_cc_time if save_duration else None,
                self._cwd,
                save_line_num,
            )
            if "ignore" == external:
                return

            if "to-files" == external:
                for repo in sorted(self._ext_ccs.keys()):
                    storeJson(
                        self.Con,
                        dest_dir,
                        True,
                        self._ext_ccs[repo],
                        self._ext_cctimes[repo] if save_duration else None,
                        self._cwd,
                        save_line_num,
                        f"_ext_{repo}",
                    )
            else:
                assert "to-external" == external
                for repo in sorted(self._ext_ccs.keys()):
                    storeJson(
                        self.Con,
                        self._ext_paths[repo],
                        True,
                        self._ext_ccs[repo],
                        self._ext_cctimes[repo] if save_duration else None,
                        self._cwd,
                        save_line_num,
                    )

        else:
            if "combine-with-overridden" != external:
                self.Con.warning(
                    "Unrecognized --external=",
                    external,
                    "value. Assuming default 'combine-with-overridden'",
                )
            take_repos = {}
            output_base = os.path.realpath(os.path.join(self._cwd, "../.."))
            for repo in sorted(self._ext_ccs.keys()):
                epath = self._ext_paths[repo]
                if not epath.startswith(output_base):
                    take_repos[repo] = epath
            self.Con.info(
                len(take_repos),
                "repos:",
                take_repos,
                "are detected to be outside of standard location for external dependencies. "
                "These will be saved into a combined file.",
            )

            storeJson(
                self.Con,
                dest_dir,
                True,
                list(itertools.chain(self._new_cc, *(self._ext_ccs[r] for r in take_repos))),
                list(
                    itertools.chain(self._new_cc_time, *(self._ext_cctimes[r] for r in take_repos))
                )
                if save_duration
                else None,
                self._cwd,
                save_line_num,
            )


class BazelWrap:
    """Takes care of communicating with Bazel, including running a build command under strace and
    producing the log file.

    More precisely:
    - checks if bazel is available
    - checks if strace is available
    - if requested, runs 'bazel clean' or 'bazel clean --expunge'
    - sets up strace to log build system execution
    - runs the build system with strace

    Irrecoverable errors are reported as YacceException.
    """

    def __init__(self, Con: LoggingConsole, args: argparse.Namespace):
        """Initializes the runner."""
        # here we define only a bare minimum of vars needed to service all public methods.
        self.Con = Con

        assert args.bazel_command and args.bazel_workspace
        assert hasattr(args, "from_log") and hasattr(args, "build_cwd")
        self._bazel: str = args.bazel_command

        args.log_file = os.path.realpath(args.log_file)

        args.bazel_workspace = os.path.realpath(args.bazel_workspace)
        if not os.path.isdir(args.bazel_workspace):
            raise YacceException(
                f"Bazel workspace directory '{args.bazel_workspace}' doesn't exist.\n"
                "Consider checking the value of '--bazel_workspace' argument."
            )
        self._bazel_workspace: str = args.bazel_workspace

        self._from_log: bool = bool(args.from_log)
        self._bazel_tested: bool = False
        self._execution_root: str | None = None
        self._path: str | None = None

    def _getPath(self) -> str:
        if not self._path:
            path = os.environ.get("PATH", "")
            if not path:
                path = os.defpath
            path = self._bazel_workspace + os.pathsep + path
            self._path = path
        return self._path

    def _resolveBinaryPath(self, binary: str) -> str:
        """Resolves the path to the binary, if necessary."""
        self.Con.debug(f"Resolving absolute path to '{binary}' executable...")
        if not os.path.isabs(binary):
            path = self._getPath()
            binary_path = shutil.which(binary, path=path)
            if not binary_path:
                raise YacceException(f"Failed to find '{binary}' in PATH='{path}'")
            self.Con.debug(f"Resolved '{binary}' to '{binary_path}'")
        else:
            binary_path = binary
        if not os.path.isfile(binary_path) or not os.access(binary_path, os.X_OK):
            raise YacceException(
                f"Binary '{binary}' (as {binary_path}) doesn't exist or is not executable"
            )
        return binary_path

    def _getExecutionRoot(self) -> str:
        if not self._execution_root:
            self._checkBazel()
            try:
                self._execution_root = self._queryBazelThrow("info", "execution_root")
                # this assumes that build system was run and hence execution_root exists
                if not self._execution_root or not os.path.isdir(self._execution_root):
                    raise YacceException(
                        f"Invalid or non-existing execution_root directory '{self._execution_root}' returned by Bazel"
                    )
            except Exception as e:
                raise YacceException(f"Failed to query Bazel execution root: {e}")

        return self._execution_root

    def fixCwdAsExecutionRoot(self, args: argparse.Namespace) -> None:
        """If necessary, queries bazel for execution root and modifies args to set it as cwd."""
        assert hasattr(args, "cwd")
        if args.cwd:
            args.cwd = os.path.realpath(args.cwd)
            # only querying bazel if the build system has to be run. Pure "from_log" should be able
            # to work even on a different machine.
            if not self._from_log:
                self.Con.debug(
                    "--cwd is specified. Checking if it matches with real execution root..."
                )
                exec_root = self._getExecutionRoot()
                if args.cwd != exec_root:
                    self.Con.warning(
                        f"Specified cwd '{args.cwd}' doesn't match to Bazel execution root '{exec_root}'. "
                        "Will ignore '--cwd' specification and use what Bazel provided."
                    )
                    args.cwd = exec_root
        else:
            self.Con.debug("Querying Bazel for the execution root...")
            args.cwd = self._getExecutionRoot()
        self._execution_root = args.cwd

    def _queryBazelThrow(self, *args) -> str:
        self.Con.debug("Querying '", self._bazel, "' with args:", args)
        r = (
            subprocess.run(
                [self._bazel, *args],
                cwd=self._bazel_workspace,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,  # bazel might spew lot's of useless stuff to stderr
            )
            .stdout.decode("utf-8")
            .rstrip()
        )
        self.Con.debug("result:", r)
        return r

    def _checkBazel(self) -> None:
        if not self._bazel_tested:
            assert self._bazel and self._bazel_workspace  # doing this once
            try:
                self._bazel = self._resolveBinaryPath(self._bazel)
            except YacceException as e:
                raise YacceException(
                    f"{e}\nIf your Bazel workspace differs from current directory and you have "
                    "set '--bazel_workspace' parameter to point to it properly, either you have to "
                    "invoke a custom Bazel binary (i.e. set '--bazel_command' argument) or "
                    "better just install bazelisk."
                )
            self.Con.info("Checking if '", self._bazel, "' works in:", self._bazel_workspace)
            try:
                v = self._queryBazelThrow("--version")
                self.Con.info(
                    f"{v} is detected in directory: '{self._bazel_workspace}' using command '{self._bazel}'"
                )
                self._bazel_tested = True
            except Exception as e:
                raise YacceException(
                    f"Failed to run Bazel ('{self._bazel}') in directory: '{self._bazel_workspace}': {e}"
                )

    def runBuild(self, args: argparse.Namespace, build_system_args: list) -> None:
        self.Con.debug("Running the build...")
        assert not args.from_log
        assert not self._execution_root, "execution_root should be queried after running the build"
        assert len(build_system_args) > 0

        assert args.log_file and args.keep_log

        if args.build_cwd:
            args.build_cwd = os.path.realpath(args.build_cwd)
            if not os.path.isdir(args.build_cwd):
                raise YacceException(
                    f"Build system working directory '{args.build_cwd}' doesn't exist.\n"
                    "Consider checking value of --build_cwd argument."
                )
        else:
            args.build_cwd = self._bazel_workspace

        if args.build_shell:
            shell_path = shutil.which(args.build_shell)
            if not shell_path:
                raise YacceException(
                    f"Failed to find specified build shell '{args.build_shell}' in PATH."
                )
            args.build_shell = shell_path
            self.Con.debug(f"Using '{args.build_shell}' as a shell to run the build command.")

        assert os.path.isabs(args.log_file)
        if os.path.exists(args.log_file):
            self.Con.warning(f"Log file '{args.log_file}' already exists and will be overwritten.")
            os.remove(args.log_file)

        self._checkBazel()
        self._checkStrace()

        self._handleClean(args)

        self._runBazelWithStrace(
            args.log_file,
            args.keep_log,
            build_system_args,
            args.build_cwd,
            args.build_shell,
            args.ensure_build_succeeds,
        )

    def _handleClean(self, args: argparse.Namespace) -> None:
        if args.clean is None:
            ans = input(
                "\nATTENTION!\n\n'--clean' argument wasn't specified. To gather all build system "
                "information, yacce has to supervise the complete build process therefore it's "
                "highly recommended to run a 'bazel clean' command before the build to ensure all "
                "files are recompiled and hence traced.\n"
                "Do you authorize yacce to run 'bazel clean' now? [Y/n]: "
            )
            if ans.lower() == "n":
                args.clean = "never"
            else:
                args.clean = "always"

        if args.clean in ("always", "expunge"):
            self._bazelClean(args.clean == "expunge")
        else:
            self.Con.debug("Skipping 'bazel clean' as requested.")

    def _checkStrace(self) -> None:
        assert self._bazel_workspace
        try:
            self._strace = self._resolveBinaryPath("strace")
        except YacceException as e:
            raise YacceException(
                f"{e}\nProbably the easiest way to install it is via your package manager, such as "
                "'apt install strace' or similar."
            )
        self.Con.info("Checking if '", self._strace, "' is available in:", self._bazel_workspace)
        try:
            run_res = subprocess.run(
                [self._strace, "--version"],
                cwd=self._bazel_workspace,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            stderr = run_res.stderr.decode("utf-8").rstrip()
            if stderr:
                self.Con.warning(
                    "'strace --version' produced output to stderr which is unexpected for return code 0. It might not work properly. STDERR =",
                    stderr,
                )
            stdout = run_res.stdout.decode("utf-8").rstrip()
            self.Con.info(f"Using {stdout}")
        except Exception as e:
            raise YacceException("Failed to check 'strace' utility version: {e}")

    def _bazelClean(self, expunge: bool) -> None:
        self.Con.info(f"Cleaning the build{' with --expunge' if expunge else ''}...")
        args = ("clean",) + (("--expunge",) if expunge else ())
        try:
            self._queryBazelThrow(*args)
            self.Con.debug("Successfully cleaned the build")
        except Exception as e:
            raise YacceException(f"Failed to '{self._bazel} {' '.join(args)}': {e}")

    def _getBazelServerPid(self) -> int:
        self.Con.debug("Querying Bazel's server PID...")
        try:
            pid_str = None
            pid_str = self._queryBazelThrow("info", "server_pid")
            pid = int(pid_str)
            if pid <= 0:
                raise YacceException(f"Invalid Bazel server PID '{pid}' returned")
            self.Con.debug(f"Bazel server PID is {pid}")
            return pid
        except ValueError:
            raise YacceException(f"Non-integer Bazel server PID '{pid_str}' was returned")
        except Exception as e:
            raise YacceException(f"Failed to query Bazel server PID: {e}")

    def _launchStrace(self, server_pid: int, log_file: str) -> subprocess.Popen:
        try:
            strace_cmd = [
                self._strace,
                "--follow-forks",
                "--trace=execve,execveat,exit",
                "--status=successful",
                "--string-limit=8192",
                "--absolute-timestamps=format:unix,precision:us",
                f"--attach={server_pid}",
                f"--output={log_file}",
            ]
            self.Con.debug(
                f"Launching strace on server_pid {server_pid}, logging to '{log_file}' with a command:",
                strace_cmd,
            )

            strace_proc = subprocess.Popen(
                strace_cmd,
                cwd=self._bazel_workspace,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                # DEVNULL,  # unfortunately strace outputs everything to stderr
            )
            retcode = strace_proc.poll()
            if retcode is not None:
                if strace_proc.stderr:
                    stderr = strace_proc.stderr.read().decode("utf-8").rstrip()
                else:
                    stderr = "<unknown>"
                raise YacceException(
                    f"Failed to launch strace on PID {server_pid}, it exited immediately with code {retcode}. STDERR='{stderr}'"
                )

            self.Con.debug(f"Strace launched with PID {strace_proc.pid}")
            return strace_proc
        except Exception as e:
            raise YacceException(f"Failed to launch strace on PID {server_pid}: {e}")

    def _runBazelWithStrace(
        self,
        log_file: str,
        keep_log: str,
        build_system_args: list,
        build_cwd: str,
        build_shell: str,
        ensure_build_succeeds: bool,
    ) -> None:
        assert log_file and build_system_args and build_cwd and build_shell
        server_pid = self._getBazelServerPid()

        shown_begin = False
        with self._launchStrace(server_pid, log_file) as strace:
            try:
                build_system_str = " ".join([shlex.quote(s) for s in build_system_args])
                self.Con.info(
                    "Running build from shell '",
                    build_shell,
                    ", directory '",
                    build_cwd,
                    "' command '",
                    build_system_str,
                    "'",
                )
                retcode = -100500
                if "linux" == sys.platform:
                    self.Con.yacce_end()
                    cwd = os.getcwd()
                    os.chdir(build_cwd)
                    try:
                        retcode = pty.spawn([build_shell, "-c", build_system_str])
                    except KeyboardInterrupt:
                        self.Con.yacce_begin()
                        shown_begin = True
                        self.Con.warning(
                            "Build interrupted by user. If you want to abort yacce too, please Ctrl-C again."
                        )
                        retcode = -1
                    finally:
                        os.chdir(cwd)
                else:
                    self.Con.yacce_end()
                    try:
                        retcode = subprocess.run(
                            build_system_args,
                            cwd=build_cwd,
                            check=False,
                            shell=True,
                            executable=build_shell,
                        ).returncode
                    except KeyboardInterrupt:
                        self.Con.yacce_begin()
                        shown_begin = True
                        self.Con.warning(
                            "Build interrupted by user. If you want to abort yacce too, please Ctrl-C again."
                        )
                        retcode = -1

                if not shown_begin:
                    self.Con.yacce_begin()
                    shown_begin = True

                if retcode != 0:
                    msg = f"Build command exited with code {retcode}."
                    if ensure_build_succeeds:
                        raise YacceException(
                            f"{msg} Yacce was requested to ensure build succeeds, so aborting. "
                            "If you want to proceed anyway, don't set --ensure_build_succeeds argument."
                        )
                    self.Con.warning(
                        f"{msg} Processing the log anyway even though the log might be incomplete "
                        "or corrupted. To enforce yacce failure if build fails, set --ensure_build_succeeds argument."
                    )

            except Exception as e:
                if not shown_begin:
                    self.Con.yacce_begin()
                    shown_begin = True
                raise YacceException(f"Failed to run the build command: {e}")

            finally:
                if not shown_begin:
                    self.Con.yacce_begin()
                    shown_begin = True
                self.Con.debug("Stopping strace.")
                strace.send_signal(signal.SIGINT)

                self.Con.debug("Waiting for completion...")
                # _, stderr = strace.communicate()
                strace.wait()
                self.Con.debug("Strace finished.")
                """
                if stderr:
                    stderr = stderr.decode("utf-8").rstrip()
                    self.Con.warning(
                        "'strace' produced output to stderr which might indicate problems. STDERR =",
                        stderr,
                    )"""

                if "never" == keep_log:
                    if os.path.exists(log_file):
                        self.Con.info(
                            f"Removing log file '{log_file}' as requested by --keep_log=never."
                        )
                        os.remove(log_file)


def mode_bazel(Con: LoggingConsole, args: argparse.Namespace, unparsed_args: list) -> int:
    args, build_system_args = _getArgs(Con, args, unparsed_args)

    Con.debug("bazel mode args: ", args)
    Con.debug("build_system_args:", build_system_args)
    Con.cleanNumErrors()

    bzl = BazelWrap(Con, args)

    if args.from_log:
        if build_system_args:
            raise YacceException(
                "Mode '--from_log' isn't intended for running a build system. Use a live mode instead. "
                "(see 'yacce bazel --help' for details)"
            )
    else:
        bzl.runBuild(args, build_system_args)

    # Only after finishing the build we could query bazel properties. Updating args.cwd from bazel
    bzl.fixCwdAsExecutionRoot(args)

    p = BazelParser(Con, args)

    # TODO proper handling of args.external
    dest_dir = (
        args.dest_dir
        if hasattr(args, "dest_dir") and args.dest_dir
        else os.path.dirname(args.log_file)
    )
    p.storeJsons(dest_dir, args.external, args.save_duration, args.save_line_num)

    # 'never' has already been processed and 'always' basically mean 'forget it'
    if not args.from_log and args.keep_log == "if_errors":
        if Con.getNumErrors() > 0:
            Con.info(
                "There were",
                Con.getNumErrors(),
                "reported during the run, hence according to --keep_log=if_errors, leaving the log file '",
                args.log_file,
                "' in place",
            )
        else:
            if os.path.exists(args.log_file):
                Con.info("No errors were detected, hence removing log file '", args.log_file, "'")
                os.remove(args.log_file)

    return 0
